"""
analysis is a wrapper for coucal which has its own unit tests.
Here we check that the modules and methods exist.
"""

from types import ModuleType

import cads_toolbox.analysis as cta

AGGREGATE_METHODS = [
    "daily_max",
    "daily_mean",
    "daily_min",
    "monthly_mean",
    "reduce",
    "resample",
    "rolling_reduce",
]

CLIMATE_METHODS = [
    "anomaly",
    "climatology_max",
    "climatology_mean",
    "climatology_median",
    "climatology_min",
    "climatology_percentiles",
    "climatology_quantiles",
    "climatology_std",
]


def test_aggregate() -> None:
    assert isinstance(cta.aggregate, ModuleType)

    for method in AGGREGATE_METHODS:
        assert method in dir(cta.aggregate)


def test_climate() -> None:
    assert isinstance(cta.climate, ModuleType)

    for method in CLIMATE_METHODS:
        assert method in dir(cta.climate)



import pathlib

import cacholote
import pytest


@pytest.fixture(autouse=True)
def config_cacholote(tmpdir: pathlib.Path):
    with cacholote.config.set(
        cache_db_urlpath="sqlite:///" + str(tmpdir / "cacholote.db"),
        cache_files_urlpath=str(tmpdir / "cache_files"),
    ):
        yield



import numpy as np
import pandas as pd
import pytest
import xarray as xr

import cads_toolbox as ct
from cads_toolbox import _inputs_transform

DA = xr.DataArray(
    np.arange(10) * 2.0, name="test", dims=["x"], coords={"x": np.arange(10)}
)
DS = DA.to_dataset()
ND = DA.values


def dummy_ndarray_function_typeset(value: np.ndarray):
    assert isinstance(value, np.ndarray)


def dummy_ndarray_function_signature(data):
    assert isinstance(data, np.ndarray)


def dummy_dataset_function_typeset(value: xr.Dataset):
    assert isinstance(value, xr.Dataset)


def dummy_dataset_function_signature(dataset):
    assert isinstance(dataset, xr.Dataset)


def dummy_dataarray_function_signature(dataarray):
    assert isinstance(dataarray, xr.DataArray)


def dummy_dataarray_function_typeset(value: xr.DataArray):
    assert isinstance(value, xr.DataArray)


def dummy_dataframe_function(value: pd.DataFrame):
    assert isinstance(value, pd.DataFrame)


@pytest.fixture
def request_args():
    return (
        "reanalysis-era5-single-levels",
        {
            "variable": "2m_temperature",
            "product_type": "reanalysis",
            "year": "2017",
            "month": "01",
            "day": "01",
            "time": "12:00",
            "grid": "10/10",
        },
    )


def test_ensure_iter():
    for thing in [list(), tuple(), dict()]:
        assert thing == _inputs_transform.ensure_iterable(thing)

    for thing in [
        "string",
        1,
        1.0,
        _inputs_transform,
        _inputs_transform.ensure_iterable,
    ]:
        assert [thing] == _inputs_transform.ensure_iterable(thing)


def test_transform(request_args):

    ct.config.USE_CACHE = False
    remote = ct.catalogue.retrieve(*request_args)
    transformed = _inputs_transform.transform(remote, xr.Dataset)
    assert type(transformed) == xr.Dataset

    local = remote.data
    transformed = _inputs_transform.transform(local, xr.Dataset)
    assert isinstance(transformed, xr.Dataset)


def test_transform_numpy_function_inputs(da=DA, ds=DS, nd=ND):
    transformed_numpy_function_typeset = _inputs_transform.transform_function_inputs(
        dummy_ndarray_function_typeset
    )
    transformed_numpy_function_typeset(nd)
    transformed_numpy_function_typeset(da)
    transformed_numpy_function_typeset(ds)

    transformed_numpy_function_signature = _inputs_transform.transform_function_inputs(
        dummy_ndarray_function_signature
    )
    transformed_numpy_function_signature(nd)
    transformed_numpy_function_signature(da)
    transformed_numpy_function_signature(ds)


def test_transform_dataset_function_inputs(da=DA, ds=DS):
    transformed_dataset_function_typeset = _inputs_transform.transform_function_inputs(
        dummy_dataset_function_typeset
    )
    transformed_dataset_function_typeset(da)
    transformed_dataset_function_typeset(ds)

    transformed_dataset_function_signature = (
        _inputs_transform.transform_function_inputs(dummy_dataset_function_signature)
    )
    transformed_dataset_function_signature(da)
    transformed_dataset_function_signature(ds)


def test_transform_dataarray_function_inputs(da=DA):
    transformed_dataarray_function_typeset = (
        _inputs_transform.transform_function_inputs(dummy_dataarray_function_typeset)
    )
    transformed_dataarray_function_typeset(da)
    # transformed_dataarray_function_typeset(ds)  # NOT YET IMPLEMENTED IN emohawk

    transformed_dataarray_function_signature = (
        _inputs_transform.transform_function_inputs(dummy_dataarray_function_signature)
    )
    transformed_dataarray_function_signature(da)
    # transformed_dataarray_function_signature(ds)  # NOT YET IMPLEMENTED IN emohawk


def test_transform_dataframe_function_inputs(da=DA, ds=DS):
    transformed_dataframe_function = _inputs_transform.transform_function_inputs(
        dummy_dataframe_function
    )
    # transformed_dataframe_function(nd)  # NOT YET IMPLEMENT IN emohawk
    transformed_dataframe_function(da)
    transformed_dataframe_function(ds)



import pathlib
from typing import Any, Dict, Tuple

import pytest

import cads_toolbox


@pytest.fixture
def era5_request_args() -> Tuple[str, Dict[str, Any]]:
    return (
        "reanalysis-era5-single-levels",
        {
            "variable": "2m_temperature",
            "product_type": "reanalysis",
            "year": "2017",
            "month": "01",
            "day": "01",
            "time": "12:00",
        },
    )


def test_uncached_download(
    tmp_path: pathlib.Path, era5_request_args: Tuple[str, Dict[str, Any]]
):
    cads_toolbox.config.USE_CACHE = False

    remote = cads_toolbox.catalogue.retrieve(*era5_request_args)
    target = tmp_path / "test.grib"

    # Download
    assert remote.download(target) == target
    assert target.stat().st_size == 2076600

    # Re-download
    previous_mtime = target.stat().st_mtime
    assert remote.download(target) == target
    assert target.stat().st_mtime != previous_mtime
    assert target.stat().st_size == 2076600


def test_cached_download(
    tmp_path: pathlib.Path, era5_request_args: Tuple[str, Dict[str, Any]]
):
    cads_toolbox.config.USE_CACHE = True

    remote = cads_toolbox.catalogue.retrieve(*era5_request_args)
    # Download to cache
    cache_file = pathlib.Path(remote.download())
    assert cache_file.stat().st_size == 2076600
    assert cache_file.parent == tmp_path / "cache_files"

    # Use cached file
    previous_mtime = cache_file.stat().st_mtime
    assert remote.download() == str(cache_file)
    assert cache_file.stat().st_mtime == previous_mtime

    # Copy from cache file
    target = tmp_path / "test.grib"
    assert remote.download(target) == target
    assert target.stat().st_size == 2076600
    assert cache_file.stat().st_mtime == previous_mtime


def test_to_xarray(
    tmp_path: pathlib.Path, era5_request_args: Tuple[str, Dict[str, Any]]
):
    pytest.importorskip("cfgrib")
    xr = pytest.importorskip("xarray")

    cads_toolbox.config.USE_CACHE = True
    remote = cads_toolbox.catalogue.retrieve(*era5_request_args)
    assert isinstance(remote.to_xarray(), xr.Dataset)


def test_to_pandas(
    tmp_path: pathlib.Path, era5_request_args: Tuple[str, Dict[str, Any]]
):
    pytest.importorskip("cfgrib")
    pd = pytest.importorskip("pandas")

    cads_toolbox.config.USE_CACHE = True
    remote = cads_toolbox.catalogue.retrieve(*era5_request_args)
    assert isinstance(remote.to_pandas(), pd.DataFrame)



"""
Module containing methods to transform the inputs of functions based on the function type setting,
common signitures or mapping defined at call time
"""
import inspect
import types
import typing as T
from functools import wraps

import emohawk
import numpy as np
import xarray as xr

from cads_toolbox.catalogue import Remote

try:
    UNION_TYPES = [T.Union, types.UnionType]
except AttributeError:
    # This sort of Union is not allowed in versions of python<3.9
    UNION_TYPES = [
        T.Union,
    ]

EMPTY_TYPES = [inspect._empty]
DEFAULT_KWARG_TYPES = {
    "dataarray": xr.DataArray,
    "dataset": xr.Dataset,
    "data": np.ndarray,
}


def ensure_iterable(input_item):
    """Ensure that an item is iterable"""
    if not isinstance(input_item, (tuple, list, dict)):
        return [input_item]
    return input_item


def transform(obj, kwarg_type):
    """Wrapper of emohawk.transform such that it also handles cads-toolbox Remote objects"""
    if isinstance(obj, Remote):
        obj = obj.data
    return emohawk.transform(obj, kwarg_type)


def transform_function_inputs(function, **kwarg_types):
    """
    Transform the inputs to a function to match the requirements.
    This function acts as a wrapper such that emohawk will handle the input arg/kwarg format.
    """

    def _wrapper(kwarg_types, *args, **kwargs):
        kwarg_types = {**DEFAULT_KWARG_TYPES, **kwarg_types}
        signature = inspect.signature(function)
        mapping = signature_mapping(signature, kwarg_types)

        for arg, name in zip(args, signature.parameters):
            kwargs[name] = arg
        # transform kwargs if necessary
        for key, value in [(k, v) for k, v in kwargs.items() if k in mapping]:
            kwarg_types = ensure_iterable(mapping[key])
            # Transform value if necessary
            if not type(value) in kwarg_types:
                for kwarg_type in kwarg_types:
                    try:
                        kwargs[key] = transform(value, kwarg_type)
                    except ValueError:
                        # Transform was not possible, move to next kwarg type.
                        # If no trnasform is possible, format is unchanged and we rely on function to raise
                        # an Error.
                        continue
                    break

        result = function(**kwargs)

        return result

    @wraps(function)
    def wrapper(*args, **kwargs):
        return _wrapper(kwarg_types, *args, **kwargs)

    return wrapper


def signature_mapping(signature, kwarg_types):
    """
    Map args and kwargs to object types, using hierarchical selection method:
    1. Type setting
    2. Explicitly defined type
    3. Do nothing
    """
    mapping = {}
    for key, parameter in signature.parameters.items():
        annotation = parameter.annotation
        if annotation not in EMPTY_TYPES:
            # 1. Use type setting from function
            if T.get_origin(annotation) in UNION_TYPES:
                kwarg_type = T.get_args(annotation)
            else:
                kwarg_type = annotation
        elif key in kwarg_types:
            # 2. Check for specifically assigned format
            kwarg_type = kwarg_types.get(key)
        else:
            # 3. Do nothing, cannot assign None, as None is a valid type
            continue
        mapping[key] = kwarg_type
    return mapping


def transform_module_inputs(module, decorator=transform_function_inputs):
    """
    Transform the inputs to all functions in a module.
    """
    for name in dir(module):
        func = getattr(module, name)
        if isinstance(func, types.FunctionType):
            setattr(module, name, decorator(func))
    return module



"""CADS Toolbox catalogue."""

# Copyright 2022, European Union.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
from typing import Any, Dict, Union

import cacholote
import cdsapi
import emohawk
import fsspec

from . import config


def _download(
    collection_id: str,
    request: Dict[str, Any],
    target: Union[str, pathlib.Path, None] = None,
) -> Union[
    fsspec.spec.AbstractBufferedFile, fsspec.implementations.local.LocalFileOpener
]:
    client = cdsapi.Client()
    path = client.retrieve(collection_id, request).download(target)
    with fsspec.open(path, "rb") as f:
        return f


class Remote:
    """
    CADS remote object class for handling data objects that are on the CADS/CDS cache volumes
    """

    def __init__(self, collection_id: str, request: Dict[str, Any]):
        """Initialise class"""
        self.collection_id = collection_id
        self.request = request

    def download(
        self, target: Union[str, pathlib.Path, None] = None
    ) -> Union[str, pathlib.Path]:
        """
        Download file with data.

        Parameters
        ----------
        target: str, optional
            Path to which to save data.

        Returns
        -------
        str: Path to which data are saved.
        """
        if config.USE_CACHE:
            with cacholote.config.set(io_delete_original=True):
                obj = cacholote.cacheable(_download)(self.collection_id, self.request)
            if target:
                obj.fs.get(obj.path, str(target))
        else:
            obj = _download(self.collection_id, self.request, target)
        return target or obj.path

    @property
    def data(self) -> emohawk.Data:
        """Object representing the requested data."""
        return emohawk.open(
            self.download(),
            exclude=["*.png", "*.json"],  # TODO: implement dataset-specific kwargs
        )

    @property
    def to_xarray(self):
        """Convert remote object to xarray.Dataset, performs a hidden cacholote download"""
        return self.data.to_xarray

    @property
    def to_pandas(self):
        """Convert remote object to pandas.DataFrame, performs a hidden cacholote download"""
        return self.data.to_pandas


def retrieve(collection_id: str, request: Dict[str, Any]) -> Remote:
    """
    Retrieve CADS data.

    Parameters
    ----------
    collection_id: str
        ID of the dataset.
    request: dict
        Parameters of the request.

    Returns
    -------
    Remote: Object with various methods to access CADS data.
    """
    return Remote(collection_id, request)



"""
Module to contains tools for analysis.
Currently ingests the tools from the following python packages:
- coucal
"""
import coucal

from cads_toolbox._inputs_transform import transform_module_inputs

aggregate = transform_module_inputs(coucal.aggregate)

climate = transform_module_inputs(coucal.climate)
